import copy
import decimal
from collections import defaultdict
from typing import Iterable
from typing import List
from typing import Type

import lib.core as constances
from lib.core.conditions import Condition
from lib.core.conditions import CONDITION_EQ as EQ
from lib.core.entities import AnnotationClassEntity
from lib.core.entities import FolderEntity
from lib.core.entities import ProjectEntity
from lib.core.entities import TeamEntity
from lib.core.entities import WorkflowEntity
from lib.core.exceptions import AppException
from lib.core.exceptions import AppValidationException
from lib.core.reporter import Reporter
from lib.core.repositories import BaseManageableRepository
from lib.core.repositories import BaseReadOnlyRepository
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.usecases.base import BaseReportableUseCae
from lib.core.usecases.base import BaseUseCase
from lib.core.usecases.base import BaseUserBasedUseCase
from requests.exceptions import RequestException
from superannotate.logger import get_default_logger

logger = get_default_logger()


class GetProjectsUseCase(BaseUseCase):
    def __init__(
            self, condition: Condition, team_id: int, projects: BaseManageableRepository,
    ):
        super().__init__()
        self._condition = condition
        self._projects = projects
        self._team_id = team_id

    def execute(self):
        if self.is_valid():
            condition = self._condition & Condition("team_id", self._team_id, EQ)
            self._response.data = self._projects.get_all(condition)
        return self._response


class GetProjectByNameUseCase(BaseUseCase):
    def __init__(
            self, name: str, team_id: int, projects: BaseManageableRepository,
    ):
        super().__init__()
        self._name = name
        self._projects = projects
        self._team_id = team_id

    def execute(self):
        if self.is_valid():
            condition = Condition("name", self._name, EQ) & Condition(
                "team_id", self._team_id, EQ
            )
            projects = self._projects.get_all(condition)
            if not projects:
                self._response.errors = AppException("Project not found.")
            else:
                project = next(
                    (project for project in projects if project.name == self._name),
                    None,
                )
                if not project:
                    self._response.errors = AppException("Project not found")
                self._response.data = project
        return self._response


class GetProjectMetaDataUseCase(BaseUseCase):
    def __init__(
            self,
            project: ProjectEntity,
            service: SuperannotateServiceProvider,
            annotation_classes: BaseManageableRepository,
            settings: BaseManageableRepository,
            workflows: BaseManageableRepository,
            projects: BaseManageableRepository,
            include_annotation_classes: bool,
            include_settings: bool,
            include_workflow: bool,
            include_contributors: bool,
            include_complete_image_count: bool,
    ):
        super().__init__()
        self._project = project
        self._service = service

        self._annotation_classes = annotation_classes
        self._settings = settings
        self._workflows = workflows
        self._projects = projects

        self._include_annotation_classes = include_annotation_classes
        self._include_settings = include_settings
        self._include_workflow = include_workflow
        self._include_contributors = include_contributors
        self._include_complete_image_count = include_complete_image_count

    @property
    def annotation_classes_use_case(self):
        return GetAnnotationClassesUseCase(classes=self._annotation_classes)

    @property
    def settings_use_case(self):
        return GetSettingsUseCase(settings=self._settings)

    @property
    def work_flow_use_case(self):
        return GetWorkflowsUseCase(
            project=self._project,
            workflows=self._workflows,
            annotation_classes=self._annotation_classes,
        )

    def execute(self):
        data = {}
        project = self._projects.get_one(
            uuid=self._project.id, team_id=self._project.team_id
        )
        data["project"] = project
        if self._include_complete_image_count:
            completed_images_data = self._service.bulk_get_folders(
                self._project.team_id, [project.id]
            )
            root_completed_count = 0
            total_completed_count = 0
            for i in completed_images_data["data"]:
                total_completed_count += i["completedCount"]
                if i["is_root"]:
                    root_completed_count = i["completedCount"]

            project.root_folder_completed_images_count = root_completed_count
            project.completed_images_count = total_completed_count

        if self._include_annotation_classes:
            data["classes"] = self.annotation_classes_use_case.execute().data

        if self._include_settings:
            data["project"].settings = self.settings_use_case.execute().data

        if self._include_workflow:
            data["workflows"] = self.work_flow_use_case.execute().data

        if self._include_contributors:
            data["contributors"] = project.users
        else:
            project.users = []

        self._response.data = data
        return self._response


class CreateProjectUseCase(BaseUseCase):
    def __init__(
            self,
            project: ProjectEntity,
            projects: BaseManageableRepository,
            backend_service_provider: SuperannotateServiceProvider,
            annotation_classes_repo: Type[BaseManageableRepository],
            workflows_repo: Type[BaseManageableRepository],
            workflows: Iterable[WorkflowEntity] = None,
            classes: List[AnnotationClassEntity] = None,
    ):

        super().__init__()
        self._project = project
        self._projects = projects
        self._annotation_classes_repo = annotation_classes_repo
        self._workflows_repo = workflows_repo
        self._workflows = workflows
        self._classes = classes
        self._backend_service = backend_service_provider

    def validate_settings(self):
        for setting in self._project.settings[:]:
            if setting.attribute == "WorkflowType":
                self._project.settings.remove(setting)
            if setting.attribute == "ImageQuality" and isinstance(setting.value, str):
                setting.value = constances.ImageQuality.get_value(setting.value)
            elif setting.attribute == "FrameRate":
                if not self._project.type == constances.ProjectType.VIDEO.value:
                    raise AppValidationException("FrameRate is available only for Video projects")
                if isinstance(setting.value, (float, int)):
                    if not (0.0001 < setting.value < 120) or decimal.Decimal(
                            str(setting.value)).as_tuple().exponent < -3:
                        raise AppValidationException("The FrameRate value range is between 0.001 - 120")
                else:
                    raise AppValidationException("The FrameRate value should be float")

    def validate_project_name(self):
        if (
                len(
                    set(self._project.name).intersection(
                        constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                    )
                )
                > 0
        ):
            self._project.name = "".join(
                "_"
                if char in constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                else char
                for char in self._project.name
            )
            logger.warning(
                "New folder name has special characters. Special characters will be replaced by underscores."
            )
        condition = Condition("name", self._project.name, EQ) & Condition(
            "team_id", self._project.team_id, EQ
        )
        for project in self._projects.get_all(condition):
            if project.name == self._project.name:
                logger.error("There are duplicated names.")
                raise AppValidationException(
                    f"Project name {self._project.name} is not unique. "
                    f"To use SDK please make project names unique."
                )

    def execute(self):
        if self.is_valid():
            # new projects can only have the status of NotStarted
            self._project.status = constances.ProjectStatus.NotStarted.value
            entity = self._projects.insert(self._project)
            # create project doesn't store attachment data so need to update
            instructions_link = self._project.instructions_link
            if instructions_link:
                entity.instructions_link = instructions_link
                self._projects.update(entity)
            self._response.data = entity
            data = {}
            # TODO delete
            # if self._settings:
            #     settings_repo = self._settings_repo(self._backend_service, entity)
            #     for setting in self._settings:
            #         for new_setting in settings_repo.get_all():
            #             if new_setting.attribute == setting.attribute:
            #                 setting_copy = copy.copy(setting)
            #                 setting_copy.id = new_setting.id
            #                 setting_copy.project_id = entity.uuid
            #                 settings_repo.update(setting_copy)
            #     data["settings"] = self._settings
            annotation_classes_mapping = {}
            if self._classes:
                annotation_repo = self._annotation_classes_repo(
                    self._backend_service, entity
                )
                for annotation_class in self._classes:
                    annotation_classes_mapping[
                        annotation_class.id
                    ] = annotation_repo.insert(annotation_class)
                self._response.data.classes = self._classes
            if self._workflows:
                set_workflow_use_case = SetWorkflowUseCase(
                    service=self._backend_service,
                    annotation_classes_repo=self._annotation_classes_repo(
                        self._backend_service, entity
                    ),
                    workflow_repo=self._workflows_repo(self._backend_service, entity),
                    steps=self._workflows,
                    project=entity,
                )
                set_workflow_response = set_workflow_use_case.execute()
                if set_workflow_response.errors:
                    self._response.errors = set_workflow_response.errors
                data["workflows"] = self._workflows

            logger.info(
                "Created project %s (ID %s) with type %s",
                self._response.data.name,
                self._response.data.id,
                constances.ProjectType.get_name(self._response.data.type),
            )
        return self._response


class DeleteProjectUseCase(BaseUseCase):
    def __init__(
            self, project_name: str, team_id: int, projects: BaseManageableRepository,
    ):

        super().__init__()
        self._project_name = project_name
        self._team_id = team_id
        self._projects = projects

    def execute(self):
        use_case = GetProjectByNameUseCase(
            name=self._project_name, team_id=self._team_id, projects=self._projects,
        )
        project_response = use_case.execute()
        if project_response.data:
            is_deleted = self._projects.delete(project_response.data)
            if is_deleted:
                logger.info("Successfully deleted project ")
            else:
                raise AppException("Couldn't delete project")


class UpdateProjectUseCase(BaseUseCase):
    def __init__(
            self,
            project: ProjectEntity,
            project_data: dict,
            projects: BaseManageableRepository,
    ):

        super().__init__()
        self._project = project
        self._project_data = project_data
        self._projects = projects

    def validate_project_name(self):
        if self._project_data.get("name"):
            if (
                    len(
                        set(self._project_data["name"]).intersection(
                            constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                        )
                    )
                    > 0
            ):
                self._project_data["name"] = "".join(
                    "_"
                    if char in constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                    else char
                    for char in self._project_data["name"]
                )
                logger.warning(
                    "New folder name has special characters. Special characters will be replaced by underscores."
                )
            condition = Condition("name", self._project_data["name"], EQ) & Condition(
                "team_id", self._project.team_id, EQ
            )
            for project in self._projects.get_all(condition):
                if project.name == self._project_data["name"]:
                    logger.error("There are duplicated names.")
                    raise AppValidationException(
                        f"Project name {self._project_data['name']} is not unique. "
                        f"To use SDK please make project names unique."
                    )

    def execute(self):
        if self.is_valid():
            for field, value in self._project_data.items():
                setattr(self._project, field, value)
            new_project = self._projects.update(self._project)
            self._response.data = new_project
        return self._response


class CloneProjectUseCase(BaseReportableUseCae):
    def __init__(
            self,
            reporter: Reporter,
            project: ProjectEntity,
            project_to_create: ProjectEntity,
            projects: BaseManageableRepository,
            settings_repo: Type[BaseManageableRepository],
            workflows_repo: Type[BaseManageableRepository],
            annotation_classes_repo: Type[BaseManageableRepository],
            backend_service_provider: SuperannotateServiceProvider,
            include_annotation_classes: bool = True,
            include_settings: bool = True,
            include_workflow: bool = True,
            include_contributors: bool = False,
    ):
        super().__init__(reporter)
        self._project = project
        self._project_to_create = project_to_create
        self._projects = projects
        self._settings_repo = settings_repo
        self._workflows_repo = workflows_repo
        self._annotation_classes_repo = annotation_classes_repo
        self._backend_service = backend_service_provider
        self._include_annotation_classes = include_annotation_classes
        self._include_settings = include_settings
        self._include_workflow = include_workflow
        self._include_contributors = include_contributors

    @property
    def annotation_classes(self):
        return self._annotation_classes_repo(self._backend_service, self._project)

    @property
    def settings(self):
        return self._settings_repo(self._backend_service, self._project)

    @property
    def workflows(self):
        return self._workflows_repo(self._backend_service, self._project)

    def validate_project_name(self):
        if self._project_to_create.name:
            if (
                    len(
                        set(self._project_to_create.name).intersection(
                            constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                        )
                    )
                    > 0
            ):
                self._project_to_create.name = "".join(
                    "_"
                    if char in constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                    else char
                    for char in self._project_to_create.name
                )
                logger.warning(
                    "New folder name has special characters. Special characters will be replaced by underscores."
                )
            condition = Condition("name", self._project_to_create.name, EQ) & Condition(
                "team_id", self._project.team_id, EQ
            )
            for project in self._projects.get_all(condition):
                if project.name == self._project_to_create.name:
                    logger.error("There are duplicated names.")
                    raise AppValidationException(
                        f"Project name {self._project_to_create.name} is not unique. "
                        f"To use SDK please make project names unique."
                    )

    def get_annotation_classes_repo(self, project: ProjectEntity):
        return self._annotation_classes_repo(self._backend_service, project)

    def _copy_annotation_classes(
            self, annotation_classes_entity_mapping: dict, project: ProjectEntity
    ):
        annotation_classes = self.annotation_classes.get_all()
        for annotation_class in annotation_classes:
            annotation_class_copy = copy.copy(annotation_class)
            annotation_classes_entity_mapping[
                annotation_class.id
            ] = self.get_annotation_classes_repo(project).insert(annotation_class_copy)

    def _copy_include_contributors(self, to_project: ProjectEntity):
        from_project = self._projects.get_one(
            uuid=self._project.id, team_id=self._project.team_id
        )
        users = []
        for user in from_project.users:
            users.append(
                {"user_id": user.get("user_id"), "user_role": user.get("user_role")}
            )

        for user in from_project.unverified_users:
            users.append(
                {"user_id": user.get("email"), "user_role": user.get("user_role")}
            )
        if users:
            self._backend_service.share_project_bulk(
                to_project.id, to_project.team_id, users
            )

    def _copy_settings(self, to_project: ProjectEntity):
        new_settings = self._settings_repo(self._backend_service, to_project)
        for setting in self.settings.get_all():
            for new_setting in new_settings.get_all():
                if new_setting.attribute == setting.attribute:
                    setting_copy = copy.copy(setting)
                    setting_copy.id = new_setting.id
                    setting_copy.project_id = to_project.id
                    new_settings.update(setting_copy)

    def _copy_workflow(
            self, annotation_classes_entity_mapping: dict, to_project: ProjectEntity
    ):
        new_workflows = self._workflows_repo(self._backend_service, to_project)
        for workflow in self.workflows.get_all():
            existing_workflow_ids = list(map(lambda i: i.uuid, new_workflows.get_all()))
            workflow_data = copy.copy(workflow)
            workflow_data.project_id = to_project.id
            workflow_data.class_id = annotation_classes_entity_mapping[
                workflow.class_id
            ].id
            new_workflows.insert(workflow_data)
            workflows = new_workflows.get_all()
            new_workflow = next(
                (
                    work_flow
                    for work_flow in workflows
                    if work_flow.uuid not in existing_workflow_ids
                ),
                None,
            )
            workflow_attributes = []
            for attribute in workflow_data.attribute:
                for annotation_attribute in annotation_classes_entity_mapping[
                    workflow.class_id
                ].attribute_groups:
                    if (
                            attribute["attribute"]["attribute_group"]["name"]
                            == annotation_attribute.name
                    ):
                        for (
                                annotation_attribute_value
                        ) in annotation_attribute.attributes:
                            if (
                                    annotation_attribute_value.name
                                    == attribute["attribute"]["name"]
                            ):
                                workflow_attributes.append(
                                    {
                                        "workflow_id": new_workflow.uuid,
                                        "attribute_id": annotation_attribute_value.id,
                                    }
                                )
                                break
            if workflow_attributes:
                self._backend_service.set_project_workflow_attributes_bulk(
                    project_id=to_project.id,
                    team_id=to_project.team_id,
                    attributes=workflow_attributes,
                )

    def execute(self):
        if self.is_valid():
            if self._project_to_create.type in (
                    constances.ProjectType.PIXEL.value,
                    constances.ProjectType.VECTOR.value,
            ):
                self._project_to_create.upload_state = (
                    constances.UploadState.INITIAL.value
                )
            project = self._projects.insert(self._project_to_create)
            self.reporter.log_info(
                f"Created project {self._project_to_create.name} with type"
                f" {constances.ProjectType.get_name(self._project_to_create.type)}."
            )
            annotation_classes_entity_mapping = defaultdict(AnnotationClassEntity)
            annotation_classes_created = False
            if self._include_annotation_classes:
                self.reporter.log_info(
                    f"Cloning annotation classes from {self._project.name} to {self._project_to_create.name}."
                )
                try:
                    self._copy_annotation_classes(
                        annotation_classes_entity_mapping, project
                    )
                    annotation_classes_created = True
                except (AppException, RequestException) as e:
                    self.reporter.log_warning(
                        f"Failed to clone annotation classes from {self._project.name} to {self._project_to_create.name}."
                    )
                    self.reporter.log_debug(str(e), exc_info=True)

            if self._include_settings:
                self.reporter.log_info(
                    f"Cloning settings from {self._project.name} to {self._project_to_create.name}."
                )
                try:
                    self._copy_settings(project)
                except (AppException, RequestException) as e:
                    self.reporter.log_warning(
                        f"Failed to clone settings from {self._project.name} to {self._project_to_create.name}."
                    )
                    self.reporter.log_debug(str(e), exc_info=True)

            if self._include_workflow:
                if self._project.type in (
                        constances.ProjectType.DOCUMENT.value,
                        constances.ProjectType.VIDEO.value,
                ):
                    self.reporter.log_warning(
                        "Workflow copy is deprecated for "
                        f"{constances.ProjectType.get_name(self._project_to_create.type)} projects."
                    )
                elif not annotation_classes_created:
                    self.reporter.log_info(
                        f"Skipping the workflow clone from {self._project.name} to {self._project_to_create.name}."
                    )
                else:
                    self.reporter.log_info(
                        f"Cloning workflow from {self._project.name} to {self._project_to_create.name}."
                    )
                    try:
                        self._copy_workflow(annotation_classes_entity_mapping, project)
                    except (AppException, RequestException) as e:
                        self.reporter.log_warning(
                            f"Failed to workflow from {self._project.name} to {self._project_to_create.name}."
                        )
                        self.reporter.log_debug(str(e), exc_info=True)
            if self._include_contributors:
                self.reporter.log_info(
                    f"Cloning contributors from {self._project.name} to {self._project_to_create.name}."
                )
                try:
                    self._copy_include_contributors(project)
                except (AppException, RequestException) as e:
                    self.reporter.log_warning(
                        f"Failed to clone contributors from {self._project.name} to {self._project_to_create.name}."
                    )
                    self.reporter.log_debug(str(e), exc_info=True)
            self._response.data = self._projects.get_one(
                uuid=project.id, team_id=project.team_id
            )

        return self._response


class ShareProjectUseCase(BaseUseCase):
    def __init__(
            self,
            service: SuperannotateServiceProvider,
            project_entity: ProjectEntity,
            user_id: str,
            user_role: str,
    ):
        super().__init__()
        self._service = service
        self._project_entity = project_entity
        self._user_id = user_id
        self._user_role = user_role

    @property
    def user_role(self):
        return constances.UserRole.get_value(self._user_role)

    def execute(self):
        self._response.data = self._service.share_project_bulk(
            team_id=self._project_entity.team_id,
            project_id=self._project_entity.id,
            users=[{"user_id": self._user_id, "user_role": self.user_role}],
        )
        if not self._response.errors:
            logger.info(
                f"Shared project {self._project_entity.name} with user {self._user_id} and role {constances.UserRole.get_name(self.user_role)}"
            )
        return self._response


class UnShareProjectUseCase(BaseUseCase):
    def __init__(
            self,
            service: SuperannotateServiceProvider,
            project_entity: ProjectEntity,
            user_id: str,
    ):
        super().__init__()
        self._service = service
        self._project_entity = project_entity
        self._user_id = user_id

    def execute(self):
        self._response.data = self._service.un_share_project(
            team_id=self._project_entity.team_id,
            project_id=self._project_entity.id,
            user_id=self._user_id,
        )
        logger.info(
            f"Unshared project {self._project_entity.name} from user ID {self._user_id}"
        )
        return self._response


class GetSettingsUseCase(BaseUseCase):
    def __init__(self, settings: BaseManageableRepository):
        super().__init__()
        self._settings = settings

    def execute(self):
        self._response.data = self._settings.get_all()
        return self._response


class GetWorkflowsUseCase(BaseUseCase):
    def __init__(
            self,
            project: ProjectEntity,
            annotation_classes: BaseReadOnlyRepository,
            workflows: BaseManageableRepository,
            fill_classes=True,
    ):
        super().__init__()
        self._project = project
        self._workflows = workflows
        self._annotation_classes = annotation_classes
        self._fill_classes = fill_classes

    def validate_project_type(self):
        if self._project.type in constances.LIMITED_FUNCTIONS:
            raise AppValidationException(
                constances.LIMITED_FUNCTIONS[self._project.type]
            )

    def execute(self):
        if self.is_valid():
            data = []
            workflows = self._workflows.get_all()
            for workflow in workflows:
                workflow_data = workflow.to_dict()
                if self._fill_classes:
                    annotation_classes = self._annotation_classes.get_all()
                    for annotation_class in annotation_classes:
                        if annotation_class.id == workflow.class_id:
                            workflow_data["className"] = annotation_class.name
                            break
                data.append(workflow_data)
            self._response.data = data
        return self._response


class GetAnnotationClassesUseCase(BaseUseCase):
    def __init__(
            self, classes: BaseManageableRepository, condition: Condition = None,
    ):
        super().__init__()
        self._classes = classes
        self._condition = condition

    def execute(self):
        self._response.data = self._classes.get_all(condition=self._condition)
        return self._response


class UpdateSettingsUseCase(BaseUseCase):
    def __init__(
            self,
            projects: BaseReadOnlyRepository,
            settings: BaseManageableRepository,
            to_update: List,
            backend_service_provider: SuperannotateServiceProvider,
            project_id: int,
            team_id: int,
    ):
        super().__init__()
        self._projects = projects
        self._settings = settings
        self._to_update = to_update
        self._backend_service_provider = backend_service_provider
        self._project_id = project_id
        self._team_id = team_id

    def validate_image_quality(self):
        for setting in self._to_update:
            if setting["attribute"].lower() == "imagequality" and isinstance(
                    setting["value"], str
            ):
                setting["value"] = constances.ImageQuality.get_value(setting["value"])
                return

    def validate_project_type(self):
        project = self._projects.get_one(uuid=self._project_id, team_id=self._team_id)
        for attribute in self._to_update:
            if attribute.get(
                    "attribute", ""
            ) == "ImageQuality" and project.type in [
                constances.ProjectType.VIDEO.value,
                constances.ProjectType.DOCUMENT.value,
            ]:
                raise AppValidationException(
                    constances.DEPRICATED_DOCUMENT_VIDEO_MESSAGE
                )

    def execute(self):
        if self.is_valid():
            old_settings = self._settings.get_all()
            attr_id_mapping = {}
            for setting in old_settings:
                attr_id_mapping[setting.attribute] = setting.id

            new_settings_to_update = []
            for new_setting in self._to_update:
                new_settings_to_update.append(
                    {
                        "id": attr_id_mapping[new_setting["attribute"]],
                        "attribute": new_setting["attribute"],
                        "value": new_setting["value"],
                    }
                )

            self._response.data = self._backend_service_provider.set_project_settings(
                project_id=self._project_id,
                team_id=self._team_id,
                data=new_settings_to_update,
            )
        return self._response


class GetProjectImageCountUseCase(BaseUseCase):
    def __init__(
            self,
            service: SuperannotateServiceProvider,
            project: ProjectEntity,
            folder: FolderEntity,
            with_all_sub_folders: bool = False,
    ):
        super().__init__()
        self._service = service
        self._project = project
        self._folder = folder
        self._with_all_sub_folders = with_all_sub_folders

    def validate_user_input(self):
        if not self._folder.name == "root" and self._with_all_sub_folders:
            raise AppValidationException("The folder does not contain any sub-folders.")

    def validate_project_type(self):
        if self._project.type in constances.LIMITED_FUNCTIONS:
            raise AppValidationException(
                constances.LIMITED_FUNCTIONS[self._project.type]
            )

    def execute(self):
        if self.is_valid():
            data = self._service.get_project_images_count(
                project_id=self._project.id, team_id=self._project.team_id
            )
            count = 0
            if self._folder.name == "root":
                count += data["images"]["count"]
                if self._with_all_sub_folders:
                    for i in data["folders"]["data"]:
                        count += i["imagesCount"]
            else:
                for i in data["folders"]["data"]:
                    if i["id"] == self._folder.uuid:
                        count = i["imagesCount"]

            self._response.data = count
        return self._response


class SetWorkflowUseCase(BaseUseCase):
    def __init__(
            self,
            service: SuperannotateServiceProvider,
            annotation_classes_repo: BaseManageableRepository,
            workflow_repo: BaseManageableRepository,
            steps: list,
            project: ProjectEntity,
    ):
        super().__init__()
        self._service = service
        self._annotation_classes_repo = annotation_classes_repo
        self._workflow_repo = workflow_repo
        self._steps = steps
        self._project = project

    def validate_project_type(self):
        if self._project.type in constances.LIMITED_FUNCTIONS:
            raise AppValidationException(
                constances.LIMITED_FUNCTIONS[self._project.type]
            )

    def execute(self):
        if self.is_valid():
            annotation_classes = self._annotation_classes_repo.get_all()
            annotation_classes_map = {}
            annotations_classes_attributes_map = {}
            for annotation_class in annotation_classes:
                annotation_classes_map[annotation_class.name] = annotation_class.id
                for attribute_group in annotation_class.attribute_groups:
                    for attribute in attribute_group.attributes:
                        annotations_classes_attributes_map[
                            f"{annotation_class.name}__{attribute_group.name}__{attribute.name}"
                        ] = attribute.id

            for step in [step for step in self._steps if "className" in step]:
                if step.get("id"):
                    del step["id"]
                step["class_id"] = annotation_classes_map.get(step["className"], None)
                if not step["class_id"]:
                    raise AppException(
                        "Annotation class not found in set_project_workflow."
                    )

            self._service.set_project_workflow_bulk(
                team_id=self._project.team_id,
                project_id=self._project.id,
                steps=self._steps,
            )
            existing_workflows = self._workflow_repo.get_all()
            existing_workflows_map = {}
            for workflow in existing_workflows:
                existing_workflows_map[workflow.step] = workflow.uuid

            req_data = []
            for step in self._steps:
                annotation_class_name = step["className"]
                for attribute in step["attribute"]:
                    attribute_name = attribute["attribute"]["name"]
                    attribute_group_name = attribute["attribute"]["attribute_group"][
                        "name"
                    ]
                    if not annotations_classes_attributes_map.get(
                            f"{annotation_class_name}__{attribute_group_name}__{attribute_name}",
                            None,
                    ):
                        raise AppException(
                            "Attribute group name or attribute name not found in set_project_workflow."
                        )

                    if not existing_workflows_map.get(step["step"], None):
                        raise AppException("Couldn't find step in workflow")

                    req_data.append(
                        {
                            "workflow_id": existing_workflows_map[step["step"]],
                            "attribute_id": annotations_classes_attributes_map[
                                f"{annotation_class_name}__{attribute_group_name}__{attribute_name}"
                            ],
                        }
                    )

            self._service.set_project_workflow_attributes_bulk(
                project_id=self._project.id,
                team_id=self._project.team_id,
                attributes=req_data,
            )
        return self._response


class GetTeamUseCase(BaseUseCase):
    def __init__(self, teams: BaseReadOnlyRepository, team_id: int):
        super().__init__()
        self._teams = teams
        self._team_id = team_id

    def execute(self):
        try:
            self._response.data = self._teams.get_one(self._team_id)
        except Exception:
            raise AppException("Can't get team data.") from None
        return self._response


class SearchContributorsUseCase(BaseUseCase):
    def __init__(
            self,
            backend_service_provider: SuperannotateServiceProvider,
            team_id: int,
            condition: Condition = None,
    ):
        super().__init__()
        self._backend_service = backend_service_provider
        self._team_id = team_id
        self._condition = condition

    @property
    def condition(self):
        if self._condition:
            return self._condition.build_query()

    def execute(self):
        res = self._backend_service.search_team_contributors(
            self._team_id, self.condition
        )
        self._response.data = res
        return self._response


class AddContributorsToProject(BaseUserBasedUseCase):
    """
    Returns tuple of lists (added, skipped)
    """

    def __init__(
            self,
            reporter: Reporter,
            team: TeamEntity,
            project: ProjectEntity,
            emails: list,
            role: str,
            service: SuperannotateServiceProvider,
    ):
        super().__init__(reporter, emails)
        self._team = team
        self._project = project
        self._role = role
        self._service = service

    @property
    def user_role(self):
        return constances.UserRole.get_value(self._role)

    def execute(self):
        if self.is_valid():
            team_users = set()
            project_users = {user["user_id"] for user in self._project.users}
            for user in self._team.users:
                if user.user_role > constances.UserRole.ADMIN.value:
                    team_users.add(user.email)
            # collecting pending team users which is not admin
            for user in self._team.pending_invitations:
                if user["user_role"] > constances.UserRole.ADMIN.value:
                    team_users.add(user["email"])
            # collecting pending project users which is not admin
            for user in self._project.unverified_users:
                if user["user_role"] > constances.UserRole.ADMIN.value:
                    project_users.add(user["email"])

            to_add = list(team_users.intersection(self._emails) - project_users)
            to_skip = list(set(self._emails).difference(to_add))

            if to_skip:
                self.reporter.log_warning(
                    f"Skipped {len(to_skip)}/{len(self._emails)} "
                    "contributors that are out of the team scope or already have access to the project."
                )
            if to_add:
                response = self._service.share_project_bulk(
                    team_id=self._team.uuid,
                    project_id=self._project.id,
                    users=[
                        dict(user_id=user_id, user_role=self.user_role)
                        for user_id in to_add
                    ],
                )
                if response and not response.get("invalidUsers"):
                    self.reporter.log_info(
                        f"Added {len(to_add)}/{len(self._emails)} "
                        f"contributors to the project {self._project.name} with the {self._role} role."
                    )
            self._response.data = to_add, to_skip
            return self._response


class InviteContributorsToTeam(BaseUserBasedUseCase):
    """
    Returns tuple of lists (added, skipped)
    """

    def __init__(
            self,
            reporter: Reporter,
            team: TeamEntity,
            emails: list,
            set_admin: bool,
            service: SuperannotateServiceProvider,
    ):
        super().__init__(reporter, emails)
        self._team = team
        self._set_admin = set_admin
        self._service = service

    def execute(self):
        if self.is_valid():
            team_users = {user.email for user in self._team.users}
            # collecting pending team users
            team_users.update(
                {user["email"] for user in self._team.pending_invitations}
            )

            emails = set(self._emails)

            to_skip = list(emails.intersection(team_users))
            to_add = list(emails.difference(to_skip))
            invited, failed = [], to_skip
            if to_skip:
                self.reporter.log_warning(
                    f"Found {len(to_skip)}/{len(self._emails)} existing members of the team."
                )
            if to_add:
                invited, failed = self._service.invite_contributors(
                    team_id=self._team.uuid,
                    # REMINDER UserRole.VIEWER is the contributor for the teams
                    team_role=constances.UserRole.ADMIN.value
                    if self._set_admin
                    else constances.UserRole.VIEWER.value,
                    emails=to_add,
                )
                if invited:
                    self.reporter.log_info(
                        f"Sent team {'admin' if self._set_admin else 'contributor'} invitations"
                        f" to {len(invited)}/{len(self._emails)} users."
                    )
                if failed:
                    to_skip = set(to_skip)
                    to_skip.update(set(failed))
                    self.reporter.log_info(
                        f"Skipped team {'admin' if self._set_admin else 'contributor'} "
                        f"invitations for {len(failed)}/{len(self._emails)} users."
                    )
            self._response.data = invited, list(to_skip)
            return self._response



"""
Supervisely to SA conversion method
"""
import json
import threading
from pathlib import Path

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import blue_color_generator
from ....common import hex_to_rgb
from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_pixel_instance
from ..sa_json_helper import _create_sa_json
from .supervisely_helper import _base64_to_polygon
from .supervisely_helper import _create_attribute_list

logger = get_default_logger()


def supervisely_instance_segmentation_to_sa_pixel(json_files, class_id_map, output_dir):
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_files), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()

    for json_file in json_files:
        file_name = "%s___pixel.json" % Path(json_file).stem

        json_data = json.load(open(json_file))
        sa_instances = []

        H, W = json_data["size"]["height"], json_data["size"]["width"]
        mask = np.zeros((H, W, 4))
        sa_metadata = {"name": Path(json_file).stem, "width": W, "height": H}

        hex_colors = blue_color_generator(10 * len(json_data["objects"]))
        index = 0
        for obj in json_data["objects"]:
            if "classTitle" in obj and obj["classTitle"] in class_id_map.keys():
                attributes = []
                if "tags" in obj.keys():
                    attributes = _create_attribute_list(
                        obj["tags"], obj["classTitle"], class_id_map
                    )
                    parts = []
                    if obj["geometryType"] == "bitmap":
                        segments = _base64_to_polygon(obj["bitmap"]["data"])
                        for segment in segments:
                            ppoints = [
                                x + obj["bitmap"]["origin"][0]
                                if i % 2 == 0
                                else x + obj["bitmap"]["origin"][1]
                                for i, x in enumerate(segment)
                            ]
                            bitmask = np.zeros((H, W)).astype(np.uint8)
                            pts = np.array(
                                [
                                    ppoints[2 * i : 2 * (i + 1)]
                                    for i in range(len(ppoints) // 2)
                                ],
                                dtype=np.int32,
                            )
                            cv2.fillPoly(bitmask, [pts], 1)
                            color = hex_to_rgb(hex_colors[index])
                            mask[bitmask == 1] = list(color[::-1]) + [255]
                            parts.append({"color": hex_colors[index]})
                            index += 1
                        cv2.imwrite(
                            str(
                                output_dir
                                / file_name.replace("___pixel.json", "___save.png")
                            ),
                            mask,
                        )
                        sa_obj = _create_pixel_instance(
                            parts, attributes, obj["classTitle"]
                        )
                        sa_instances.append(sa_obj)

        images_converted.append(Path(json_file).stem)
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()



import requests
from superannotate.logger import get_default_logger

logger = get_default_logger()


def image_downloader(url, file_name):
    r = requests.get(url, stream=True)
    if r.status_code == 200:
        with open(file_name, "wb") as f:
            f.write(r.content)
        return True
    return False


def _create_classes_id_map(json_data):
    classes = {}
    for d in json_data:
        if "objects" not in d["Label"].keys():
            continue

        instances = d["Label"]["objects"]
        for instance in instances:
            class_name = instance["value"]
            if class_name not in classes.keys():
                color = instance["color"]
                classes[class_name] = {"color": color, "attribute_groups": {}}

            if "classifications" in instance.keys():
                classifications = instance["classifications"]
                for classification in classifications:
                    if (
                        classification["value"]
                        not in classes[class_name]["attribute_groups"]
                    ):
                        if "answer" in classification.keys():
                            if isinstance(classification["answer"], str):
                                continue

                            classes[class_name]["attribute_groups"][
                                classification["value"]
                            ] = {"is_multiselect": 0, "attributes": []}
                            if isinstance(classification["answer"], list):
                                classes[class_name]["attribute_groups"][
                                    classification["value"]
                                ]["attributes"].append(
                                    classification["answer"][0]["value"]
                                )
                            elif isinstance(classification["answer"], dict):
                                classes[class_name]["attribute_groups"][
                                    classification["value"]
                                ]["attributes"].append(
                                    classification["answer"]["value"]
                                )

                        elif "answers" in classification.keys():
                            classes[class_name]["attribute_groups"][
                                classification["value"]
                            ] = {"is_multiselect": 1, "attributes": []}
                            for attr in classification["answers"]:
                                classes[class_name]["attribute_groups"][
                                    classification["value"]
                                ]["attributes"].append(attr["value"])

                    else:
                        if "answer" in classification.keys():
                            if isinstance(classification["answer"], dict):
                                classes[class_name]["attribute_groups"][
                                    classification["value"]
                                ]["attributes"].append(
                                    classification["answer"]["value"]
                                )
                            elif isinstance(classification["answer"], list):
                                classes[class_name]["attribute_groups"][
                                    classification["value"]
                                ]["attributes"].append(
                                    classification["answer"][0]["value"]
                                )
                        elif "answers" in classification.keys():
                            for attr in classification["answers"]:
                                classes[class_name]["attribute_groups"][
                                    classification["value"]
                                ]["attributes"].append(attr["value"])
    return classes


def _create_attributes_list(lb_attributes):
    attributes = []
    for attribute in lb_attributes:
        group_name = attribute["value"]
        if "answer" in attribute.keys() and isinstance(attribute["answer"], dict):
            attribute_name = attribute["answer"]["value"]
            attr_dict = {"name": attribute_name, "groupName": group_name}
            attributes.append(attr_dict)
        elif "answers" in attribute.keys():
            for attr in attribute["answers"]:
                attribute_name = attr["value"]
                attr_dict = {"name": attribute_name, "groupName": group_name}
                attributes.append(attr_dict)

    return attributes



"""
Module which will convert from other
annotation formats to superannotate annotation format
"""
import shutil
from pathlib import Path

from superannotate.logger import get_default_logger

from .converters.converters import Converter

logger = get_default_logger()


def _load_files(path_to_imgs, ptype, extensions):
    if extensions is None:
        extensions = ["jpg", "jpeg", "png", "tif", "tiff", "webp", "bmp"]

    logger.info(
        "All files with following extensions %s will be copied to output folder",
        extensions,
    )

    images = []
    for extension in extensions:
        rec_search = str(Path("**") / ("*." + extension))
        images_gen = Path(path_to_imgs).glob(rec_search)
        images.extend(list(images_gen))

    if not images:
        logger.warning("Images doesn't exist")

    return images


def _move_files(imgs, output_dir):
    (output_dir / "classes").mkdir(parents=True, exist_ok=True)
    output_path = output_dir

    for im in imgs:
        shutil.copy(im, output_path / Path(im).name)


def import_to_sa(args):
    """
    :param args: All arguments that will be used during convertion.
    :type args: Namespace
    """

    images = _load_files(
        args.input_dir / args.images_root, args.project_type, args.images_extensions
    )
    _move_files(images, args.output_dir)

    args.__dict__.update({"direction": "from", "export_root": args.input_dir})
    converter = Converter(args)

    converter.convert_to_sa()

    logger.info("Conversion completed")



from enum import Enum
from types import DynamicClassAttribute


class BaseTitledEnum(int, Enum):
    def __new__(cls, title, value):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.__doc__ = title
        obj._type = "titled_enum"
        return obj

    @DynamicClassAttribute
    def name(self) -> str:
        return self.__doc__

    @DynamicClassAttribute
    def value(self):
        return super().value

    @classmethod
    def get_name(cls, value):
        for enum in list(cls):
            if enum.value == value:
                return enum.__doc__

    @classmethod
    def get_value(cls, name):
        for enum in list(cls):
            if enum.__doc__.lower() == name.lower():
                return enum.value

    @classmethod
    def values(cls):
        return [enum.__doc__.lower() for enum in list(cls)]

    @classmethod
    def titles(cls):
        return [enum.__doc__ for enum in list(cls)]

    def equals(self, other: Enum):
        return self.__doc__.lower() == other.__doc__.lower()


class ProjectType(BaseTitledEnum):
    VECTOR = "Vector", 1
    PIXEL = "Pixel", 2
    VIDEO = "Video", 3
    DOCUMENT = "Document", 4


class UserRole(BaseTitledEnum):
    SUPER_ADMIN = "Superadmin", 1
    ADMIN = "Admin", 2
    ANNOTATOR = "Annotator", 3
    QA = "QA", 4
    CUSTOMER = "Customer", 5
    VIEWER = "Viewer", 6


class UploadState(Enum):
    INITIAL = 1
    BASIC = 2
    EXTERNAL = 3


class ImageQuality(BaseTitledEnum):
    ORIGINAl = "original", 100
    COMPRESSED = "compressed", 60


class ProjectStatus(BaseTitledEnum):
    Undefined = "Undefined", 0
    NotStarted = "NotStarted", 1
    InProgress = "InProgress", 2
    Completed = "Completed", 3
    OnHold = "OnHold", 4


class ExportStatus(BaseTitledEnum):
    IN_PROGRESS = "inProgress", 1
    COMPLETE = "complete", 2
    CANCELED = "canceled", 3
    ERROR = "error", 4


class AnnotationStatus(BaseTitledEnum):
    NOT_STARTED = "NotStarted", 1
    IN_PROGRESS = "InProgress", 2
    QUALITY_CHECK = "QualityCheck", 3
    RETURNED = "Returned", 4
    COMPLETED = "Completed", 5
    SKIPPED = "Skipped", 6


class ClassTypeEnum(BaseTitledEnum):
    OBJECT = "object", 1
    TAG = "tag", 2

    @classmethod
    def get_value(cls, name):
        for enum in list(cls):
            if enum.__doc__.lower() == name.lower():
                return enum.value
        return cls.OBJECT.value


class IntegrationTypeEnum(BaseTitledEnum):
    AWS = "aws", 1
    GCP = "gcp", 2
    AZURE = "azure", 3


class TrainingStatus(BaseTitledEnum):
    NOT_STARTED = "NotStarted", 1
    IN_PROGRESS = "InProgress", 2
    COMPLETED = "Completed", 3
    FAILED_BEFORE_EVALUATION = "FailedBeforeEvaluation", 4
    FAILED_AFTER_EVALUATION = "FailedAfterEvaluation", 5
    FAILED_AFTER_EVALUATION_WITH_SAVE_MODEL = "FailedAfterEvaluationWithSavedModel", 6


class SegmentationStatus(BaseTitledEnum):
    NOT_STARTED = "NotStarted", 1
    IN_PROGRESS = "InProgress", 2
    COMPLETED = "Completed", 3
    FAILED = "Failed", 4


class ProjectState(BaseTitledEnum):
    NOT_SYNCED = "Not synced", 1
    SYNCED = "Synced", 2


class TrainingTask:
    INSTANCE_SEGMENTATION_PIXEL = (
        "Instance Segmentation for Pixel Projects",
        "instance_segmentation_pixel",
    )
    INSTANCE_SEGMENTATION_VECTOR = (
        "Instance Segmentation for Vector Projects",
        "instance_segmentation_vector",
    )
    KEYPOINT_DETECTION_VECTOR = (
        "Keypoint Detection for Vector Projects",
        "keypoint_detection_vector",
    )
    OBJECT_DETECTION_VECTOR = (
        "Object Detection for Vector Projects",
        "object_detection_vector",
    )
    SEMANTIC_SEGMENTATION_PIXEL = (
        "Semantic Segmentation for Pixel Projects",
        "semantic_segmentation_pixel",
    )



"""
Googlecloud to SA conversion method
"""
import threading
from pathlib import Path

import cv2
import pandas as pd
from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance

logger = get_default_logger()


def googlecloud_to_sa_vector(path, output_dir):
    df = pd.read_csv(path, header=None)
    dir_name = path.parent

    sa_jsons = {}
    classes = []
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(df), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for _, row in df.iterrows():
        classes.append(row[2])

        file_name = row[1].split("/")[-1]
        try:
            images_converted.append(file_name)
            img = cv2.imread(str(dir_name / file_name))
            H, W, _ = img.shape
        except Exception as e:
            logger.warning("Can't open %s image.", file_name)
            images_not_converted.append(file_name)
            continue

        sa_file_name = "%s___objects.json" % Path(file_name).name

        points = (row[3] * W, row[4] * H, row[5] * W, row[8] * H)
        sa_instances = _create_vector_instance("bbox", points, {}, [], row[2])

        if sa_file_name in sa_jsons.keys():
            sa_jsons[sa_file_name]["instances"].append(sa_instances)
        else:
            sa_metadata = {"name": Path(file_name).name, "width": W, "height": H}
            sa_jsons[sa_file_name] = {
                "metadata": sa_metadata,
                "instances": [sa_instances],
            }
    finish_event.set()
    tqdm_thread.join()

    for key, value in sa_jsons.items():
        sa_json = _create_sa_json(value["instances"], value["metadata"])
        write_to_json(output_dir / key, sa_json)

    return classes



"""
Main module for input converters
"""
from argparse import Namespace
from pathlib import Path

from lib.app.exceptions import AppException
from lib.app.mixp.decorators import Trackable
from lib.core import DEPRICATED_DOCUMENT_VIDEO_MESSAGE
from lib.core import LIMITED_FUNCTIONS
from lib.core.enums import ProjectType

from .export_from_sa_conversions import export_from_sa
from .import_to_sa_conversions import import_to_sa
from .sa_conversion import degrade_json
from .sa_conversion import sa_convert_project_type
from .sa_conversion import upgrade_json

ALLOWED_TASK_TYPES = [
    "panoptic_segmentation",
    "instance_segmentation",
    "keypoint_detection",
    "object_detection",
    "vector_annotation",
]

ALLOWED_PROJECT_TYPES = ["Pixel", "Vector"]

ALLOWED_ANNOTATION_IMPORT_FORMATS = {
    "COCO": [
        ("Pixel", "panoptic_segmentation"),
        ("Pixel", "instance_segmentation"),
        ("Vector", "keypoint_detection"),
        ("Vector", "instance_segmentation"),
        ("Vector", "object_detection"),
    ],
    "VOC": [
        ("Vector", "object_detection"),
        ("Vector", "instance_segmentation"),
        ("Pixel", "instance_segmentation"),
    ],
    "LabelBox": [
        ("Vector", "object_detection"),
        ("Vector", "instance_segmentation"),
        ("Vector", "vector_annotation"),
        ("Pixel", "instance_segmentation"),
    ],
    "DataLoop": [
        ("Vector", "object_detection"),
        ("Vector", "instance_segmentation"),
        ("Vector", "vector_annotation"),
    ],
    "Supervisely": [
        ("Vector", "vector_annotation"),
        ("Vector", "object_detection"),
        ("Vector", "instance_segmentation"),
        ("Pixel", "instance_segmentation"),
        ("Vector", "keypoint_detection"),
    ],
    "VoTT": [
        ("Vector", "object_detection"),
        ("Vector", "instance_segmentation"),
        ("Vector", "vector_annotation"),
    ],
    "SageMaker": [("Pixel", "instance_segmentation"), ("Vector", "object_detection")],
    "VGG": [
        ("Vector", "object_detection"),
        ("Vector", "instance_segmentation"),
        ("Vector", "vector_annotation"),
    ],
    "GoogleCloud": [("Vector", "object_detection")],
    "YOLO": [("Vector", "object_detection")],
}

ALLOWED_ANNOTATION_EXPORT_FORMATS = {
    "COCO": [
        ("Pixel", "panoptic_segmentation"),
        ("Pixel", "instance_segmentation"),
        ("Vector", "instance_segmentation"),
        ("Vector", "keypoint_detection"),
        ("Vector", "object_detection"),
    ]
}


def _change_type(input_dir, output_dir):
    if isinstance(input_dir, str):
        input_dir = Path(input_dir)
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    return input_dir, output_dir


def _passes_type_sanity(params_info):
    for param in params_info:
        if not isinstance(param[0], param[2]):
            raise AppException(
                "'{}' should be {} type, not {}".format(
                    param[1], param[2], type(param[0])
                )
            )


def _passes_list_members_type_sanity(lists_info):
    for _list in lists_info:
        for _list_member in _list[0]:
            if not isinstance(_list_member, _list[2]):
                raise AppException(
                    "'%s' should be list of '%s', but contains '%s'"
                    % (_list[1], _list[2], type(_list_member))
                )


def _passes_value_sanity(values_info):
    for value in values_info:
        if value[0] not in value[2]:
            raise AppException(
                "'{}' should be one of the following '{}'".format(value[1], value[2])
            )


def _passes_converter_sanity(args, direction):
    converter_values = (args.project_type, args.task)
    test_passed = False
    if direction == "import":
        if converter_values in ALLOWED_ANNOTATION_IMPORT_FORMATS[args.dataset_format]:
            test_passed = True
    else:
        if converter_values in ALLOWED_ANNOTATION_EXPORT_FORMATS[args.dataset_format]:
            test_passed = True

    if not test_passed:
        raise AppException(
            "Please enter valid converter values. You can check available candidates in the documentation (https://superannotate.readthedocs.io/en/stable/index.html)."
        )


@Trackable
def export_annotation(
    input_dir,
    output_dir,
    dataset_format,
    dataset_name,
    project_type="Vector",
    task="object_detection",
):
    """
       Converts SuperAnnotate annotation formate to the other annotation formats. Currently available (project_type, task) combinations for converter
       presented below:

       ==============  ======================
                From SA to COCO
       --------------------------------------
        project_type           task
       ==============  ======================
       Pixel           panoptic_segmentation
       Pixel           instance_segmentation
       Vector          instance_segmentation
       Vector          object_detection
       Vector          keypoint_detection
       ==============  ======================

       :param input_dir: Path to the dataset folder that you want to convert.
       :type input_dir: Pathlike(str or Path)
       :param output_dir: Path to the folder, where you want to have converted dataset.
       :type output_dir: Pathlike(str or Path)
       :param dataset_format: One of the formats that are possible to convert. Available candidates are: ["COCO"]
       :type dataset_format: str
       :param dataset_name: Will be used to create json file in the output_dir.
       :type dataset_name: str
       :param project_type: SuperAnnotate project type is either 'Vector' or 'Pixel' (Default: 'Vector')
                            'Vector' project creates <image_name>___objects.json for each image.
                            'Pixel' project creates <image_name>___pixel.jsons and <image_name>___save.png annotation mask for each image.
       :type project_type: str
       :param task: Task can be one of the following: ['panoptic_segmentation', 'instance_segmentation',
                    'keypoint_detection', 'object_detection']. (Default: "object_detection").
                    'keypoint_detection' can be used to converts keypoints from/to available annotation format.
                    'panoptic_segmentation' will use panoptic mask for each image to generate bluemask for SuperAnnotate annotation format and use bluemask to generate panoptic mask for invert conversion. Panoptic masks should be in the input folder.
                    'instance_segmentation' 'Pixel' project_type converts instance masks and 'Vector' project_type generates bounding boxes and polygons from instance masks. Masks should be in the input folder if it is 'Pixel' project_type.
                    'object_detection' converts objects from/to available annotation format
       :type task: str
       """

    if project_type in [
        ProjectType.VIDEO.name,
        ProjectType.DOCUMENT.name,
    ]:
        raise AppException(LIMITED_FUNCTIONS[ProjectType.get_value(project_type)])

    params_info = [
        (input_dir, "input_dir", (str, Path)),
        (output_dir, "output_dir", (str, Path)),
        (dataset_name, "dataset_name", str),
        (dataset_format, "dataset_format", str),
        (project_type, "project_type", str),
        (task, "task", str),
    ]
    _passes_type_sanity(params_info)
    input_dir, output_dir = _change_type(input_dir, output_dir)

    values_info = [
        (project_type, "project_type", ALLOWED_PROJECT_TYPES),
        (task, "task", ALLOWED_TASK_TYPES),
        (
            dataset_format,
            "dataset_format",
            list(ALLOWED_ANNOTATION_EXPORT_FORMATS.keys()),
        ),
    ]
    _passes_value_sanity(values_info)

    args = Namespace(
        input_dir=input_dir,
        output_dir=output_dir,
        dataset_format=dataset_format,
        dataset_name=dataset_name,
        project_type=project_type,
        task=task,
    )

    _passes_converter_sanity(args, "export")

    export_from_sa(args)


@Trackable
def import_annotation(
    input_dir,
    output_dir,
    dataset_format="superannotate",
    dataset_name="",
    project_type="Vector",
    task="object_detection",
    images_root="",
    images_extensions=None,
):
    """Converts other annotation formats to SuperAnnotate annotation format. Currently available (project_type, task) combinations for converter
    presented below:

    ==============  ======================
             From COCO to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           panoptic_segmentation
    Pixel           instance_segmentation
    Vector          instance_segmentation
    Vector          object_detection
    Vector          keypoint_detection
    ==============  ======================

    ==============  ======================
             From VOC to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           instance_segmentation
    Vector          instance_segmentation
    Vector          object_detection
    ==============  ======================

    ==============  ======================
           From LabelBox to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    Vector          instance_segmentation
    Vector          vector_annotation
    Pixel           instance_segmentation
    ==============  ======================

    ==============  ======================
           From DataLoop to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    Vector          instance_segmentation
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From Supervisely to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    Vector          keypoint_detection
    Vector          vector_annotation
    Vector          instance_segmentation
    Pixel           instance_segmentation
    ==============  ======================

    ==============  ======================
           From VoTT to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          instance_segmentation
    Vector          object_detection
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From SageMaker to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           instance_segmentation
    Vector          objcet_detection
    ==============  ======================

    ==============  ======================
           From VGG to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          instance_segmentation
    Vector          object_detection
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From GoogleCloud to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    ==============  ======================

    ==============  ======================
           From YOLO to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    ==============  ======================

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: Pathlike(str or Path)
    :param output_dir: Path to the folder, where you want to have converted dataset.
    :type output_dir: Pathlike(str or Path)
    :param dataset_format: Annotation format to convert SuperAnnotate annotation format. Available candidates are: ["COCO", "VOC", "LabelBox", "DataLoop",
                        "Supervisely", 'VGG', 'YOLO', 'SageMake', 'VoTT', 'GoogleCloud']
    :type dataset_format: str
    :param dataset_name: Name of the json file in the input_dir, which should be converted.
    :type dataset_name: str
    :param project_type: SuperAnnotate project type is either 'Vector' or 'Pixel' (Default: 'Vector')
                         'Vector' project creates <image_name>___objects.json for each image.
                         'Pixel' project creates <image_name>___pixel.jsons and <image_name>___save.png annotation mask for each image.
    :type project_type: str
    :param task: Task can be one of the following: ['panoptic_segmentation', 'instance_segmentation',
                 'keypoint_detection', 'object_detection', 'vector_annotation']. (Default: "object_detection").
                 'keypoint_detection' can be used to converts keypoints from/to available annotation format.
                 'panoptic_segmentation' will use panoptic mask for each image to generate bluemask for SuperAnnotate annotation format and use bluemask to generate panoptic mask for invert conversion. Panoptic masks should be in the input folder.
                 'instance_segmentation' 'Pixel' project_type converts instance masks and 'Vector' project_type generates bounding boxes and polygons from instance masks. Masks should be in the input folder if it is 'Pixel' project_type.
                 'object_detection' converts objects from/to available annotation format
                 'vector_annotation' can be used to convert all annotations (point, ellipse, circule, cuboid and etc) to SuperAnnotate vector project.
    :param images_root: Additonal path to images directory in input_dir
    :type images_root: str
    :param image_extensions: List of image files xtensions in the images_root folder
    :type image_extensions: list

    """

    params_info = [
        (input_dir, "input_dir", (str, Path)),
        (output_dir, "output_dir", (str, Path)),
        (dataset_name, "dataset_name", str),
        (dataset_format, "dataset_format", str),
        (project_type, "project_type", str),
        (task, "task", str),
        (images_root, "images_root", str),
    ]

    if images_extensions is not None:
        params_info.append((images_extensions, "image_extensions", list))

    _passes_type_sanity(params_info)

    values_info = [
        (project_type, "project_type", ALLOWED_PROJECT_TYPES),
        (task, "task", ALLOWED_TASK_TYPES),
        (
            dataset_format,
            "dataset_format",
            list(ALLOWED_ANNOTATION_IMPORT_FORMATS.keys()),
        ),
    ]
    _passes_value_sanity(values_info)

    input_dir, output_dir = _change_type(input_dir, output_dir)
    args = Namespace(
        input_dir=input_dir,
        output_dir=output_dir,
        dataset_format=dataset_format,
        dataset_name=dataset_name,
        project_type=project_type,
        task=task,
        images_root=images_root,
        images_extensions=images_extensions,
    )

    _passes_converter_sanity(args, "import")

    import_to_sa(args)


@Trackable
def convert_project_type(input_dir, output_dir):
    """ Converts SuperAnnotate 'Vector' project type to 'Pixel' or reverse.

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: Pathlike(str or Path)
    :param output_dir: Path to the folder where you want to have converted files.
    :type output_dir: Pathlike(str or Path)

    """
    params_info = [
        (input_dir, "input_dir", (str, Path)),
        (output_dir, "output_dir", (str, Path)),
    ]
    _passes_type_sanity(params_info)
    json_paths = list(Path(str(input_dir)).glob("*.json"))
    if (
        json_paths
        and "___pixel.json" not in json_paths[0].name
        and "___objects.json" not in json_paths[0].name
    ):
        raise AppException(DEPRICATED_DOCUMENT_VIDEO_MESSAGE)

    input_dir, output_dir = _change_type(input_dir, output_dir)

    sa_convert_project_type(input_dir, output_dir)


@Trackable
def convert_json_version(input_dir, output_dir, version=2):
    """
    Converts SuperAnnotate JSON versions. Newest JSON version is 2.

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: Pathlike(str or Path)
    :param output_dir: Path to the folder, where you want to have converted dataset.
    :type output_dir: Pathlike(str or Path)
    :param version: Output version number. Currently is either 1 or 2. Default value is 2. It will upgrade version 1 to  version 2. Set 1 to degrade from version 2 to version 1.
    :type version: int

    :return: List of converted files
    :rtype: list
    """

    params_info = [
        (input_dir, "input_dir", (str, Path)),
        (output_dir, "output_dir", (str, Path)),
        (version, "version", int),
    ]
    _passes_type_sanity(params_info)
    input_dir, output_dir = _change_type(input_dir, output_dir)

    output_dir.mkdir(parents=True, exist_ok=True)
    if version == 2:
        converted_files = upgrade_json(input_dir, output_dir)
    elif version == 1:
        converted_files = degrade_json(input_dir, output_dir)
    else:
        raise AppException("'version' is either 1 or 2.")

    return converted_files



import io
import json
import os
from os.path import expanduser
from typing import List
from typing import Optional

import lib.core as constance
from lib.core.conditions import Condition
from lib.core.conditions import CONDITION_EQ as EQ
from lib.core.entities import AnnotationClassEntity
from lib.core.entities import ConfigEntity
from lib.core.entities import Entity
from lib.core.entities import FolderEntity
from lib.core.entities import ImageEntity
from lib.core.entities import IntegrationEntity
from lib.core.entities import MLModelEntity
from lib.core.entities import ProjectEntity
from lib.core.entities import S3FileEntity
from lib.core.entities import SettingEntity
from lib.core.entities import TeamEntity
from lib.core.entities import UserEntity
from lib.core.entities import WorkflowEntity
from lib.core.enums import ClassTypeEnum
from lib.core.enums import ImageQuality
from lib.core.exceptions import AppException
from lib.core.repositories import BaseManageableRepository
from lib.core.repositories import BaseProjectRelatedManageableRepository
from lib.core.repositories import BaseReadOnlyRepository
from lib.core.repositories import BaseS3Repository
from lib.infrastructure.services import SuperannotateBackendService
from pydantic import parse_obj_as


class ConfigRepository(BaseManageableRepository):
    def __init__(self, config_path: str = constance.CONFIG_FILE_LOCATION):
        self._config_path = f"{config_path}"

    @property
    def config_path(self):
        return expanduser(self._config_path)

    def _create_config(self):
        """
        Create a config file
        """
        os.makedirs(os.path.dirname(self.config_path), exist_ok=True)
        open(self.config_path, "w").close()
        return {}

    def _get_config(self) -> Optional[dict]:
        if os.path.exists(self.config_path):
            return json.load(open(self.config_path))

    def get_one(self, uuid: str) -> Optional[ConfigEntity]:
        config = self._get_config()
        if not config:
            return None
        try:
            return ConfigEntity(uuid=uuid, value=config.get(uuid))
        except KeyError:
            return None

    def get_all(self, condition: Condition = None) -> List[ConfigEntity]:
        config = self._get_config()
        if config:
            return [ConfigEntity(uuid, value) for uuid, value in config.items()]

    def insert(self, entity: ConfigEntity) -> ConfigEntity:
        config = self._get_config()
        if not config:
            config = self._create_config()
        config[entity.uuid] = entity.value
        with open(self._config_path, "w") as config_file:
            config_file.write(json.dumps(config, sort_keys=True, indent=4))
        return entity

    def update(self, entity: ConfigEntity):
        self.insert(entity)

    def delete(self, uuid: str):
        config = self._get_config()
        del config["uuid"]
        with open(constance.CONFIG_FILE_LOCATION, "rw+") as config_file:
            config_file.write(json.dumps(config, sort_keys=True, indent=4))


class ProjectRepository(BaseManageableRepository):
    def __init__(self, service: SuperannotateBackendService):
        self._service = service

    def get_one(self, uuid: int, team_id: int) -> ProjectEntity:
        return ProjectEntity(**self._service.get_project(uuid, team_id))

    def get_all(self, condition: Condition = None) -> List[ProjectEntity]:
        condition = condition.build_query() if condition else None
        return parse_obj_as(List[ProjectEntity], self._service.get_projects(condition))

    def insert(self, entity: ProjectEntity) -> ProjectEntity:
        result = self._service.create_project(entity.dict(exclude_none=True))
        if "error" in result:
            raise AppException(result["error"])
        return ProjectEntity(**result)

    def update(self, entity: ProjectEntity):
        condition = Condition("team_id", entity.team_id, EQ)
        result = self._service.update_project(
            entity.dict(exclude_none=True), query_string=condition.build_query()
        )
        return ProjectEntity(**result)

    def delete(self, entity: ProjectEntity):
        team_id = entity.team_id
        uuid = entity.id
        condition = Condition("team_id", team_id, EQ)
        return self._service.delete_project(
            uuid=uuid, query_string=condition.build_query()
        )


class S3Repository(BaseS3Repository):
    def get_one(self, uuid: str) -> S3FileEntity:
        file = io.BytesIO()
        self._resource.Object(self._bucket, uuid).download_fileobj(file)
        return S3FileEntity(uuid=uuid, data=file)

    def insert(self, entity: S3FileEntity) -> S3FileEntity:
        data = {"Key": entity.uuid, "Body": entity.data}
        if entity.metadata:
            temp = entity.metadata
            for k in temp:
                temp[k] = str(temp[k])
            data["Metadata"] = temp
        self.bucket.put_object(**data)
        return entity

    def update(self, entity: ProjectEntity):
        self._service.update_project(entity.to_dict())

    def delete(self, uuid: int):
        self._service.delete_project(uuid)

    def get_all(self, condition: Condition = None) -> List[ProjectEntity]:
        pass


class ProjectSettingsRepository(BaseProjectRelatedManageableRepository):
    def get_one(self, uuid: int) -> SettingEntity:
        raise NotImplementedError

    def delete(self, uuid: int):
        raise NotImplementedError

    def get_all(
            self, condition: Optional[Condition] = None
    ) -> List[SettingEntity]:
        data = self._service.get_project_settings(
            self._project.id, self._project.team_id
        )
        if data:
            return parse_obj_as(List[SettingEntity], data)
        return []

    def insert(self, entity: SettingEntity) -> SettingEntity:
        entity = entity.dict()
        res = self._service.set_project_settings(
            self._project.id, self._project.team_id, [entity]
        )
        return SettingEntity(**res[0])

    def update(self, entity: SettingEntity):
        if entity.attribute == "ImageQuality" and isinstance(entity.value, str):
            entity.value = ImageQuality.get_value(entity.value)
        self._service.set_project_settings(
            self._project.id, self._project.team_id, [entity.dict()]
        )
        return entity


class WorkflowRepository(BaseProjectRelatedManageableRepository):
    def get_one(self, uuid: int) -> WorkflowEntity:
        raise NotImplementedError

    def get_all(self, condition: Optional[Condition] = None) -> List[WorkflowEntity]:
        data = self._service.get_project_workflows(
            self._project.id, self._project.team_id
        )
        return [self.dict2entity(setting) for setting in data]

    def insert(self, entity: WorkflowEntity) -> WorkflowEntity:
        data = entity.to_dict()
        del data["project_id"]
        del data["attribute"]
        res = self._service.set_project_workflow(
            entity.project_id, self._project.team_id, self._drop_nones(data)
        )
        return self.dict2entity(res[0])

    def delete(self, uuid: int):
        raise NotImplementedError

    def update(self, entity: WorkflowEntity):
        raise NotImplementedError

    @staticmethod
    def dict2entity(data: dict) -> WorkflowEntity:
        return WorkflowEntity(
            uuid=data["id"],
            project_id=data["project_id"],
            class_id=data["class_id"],
            step=data["step"],
            tool=data["tool"],
            attribute=data.get("attribute", []),
        )


class FolderRepository(BaseManageableRepository):
    def __init__(self, service: SuperannotateBackendService):
        self._service = service

    def get_one(self, uuid: Condition) -> FolderEntity:
        condition = uuid.build_query()
        data = self._service.get_folder(condition)
        if data:
            return self.dict2entity(data)

    def get_all(self, condition: Optional[Condition] = None) -> List[FolderEntity]:
        condition = condition.build_query() if condition else None
        data = self._service.get_folders(condition)
        return [self.dict2entity(image) for image in data]

    def insert(self, entity: FolderEntity) -> FolderEntity:
        res = self._service.create_folder(
            project_id=entity.project_id,
            team_id=entity.team_id,
            folder_name=entity.name,
        )
        return self.dict2entity(res)

    def update(self, entity: FolderEntity):
        project_id = entity.project_id
        team_id = entity.team_id
        response = self._service.update_folder(project_id, team_id, entity.to_dict())
        if response:
            return self.dict2entity(response)

    def delete(self, entity: FolderEntity):
        return self._service.delete_folders(
            entity.project_id, entity.team_id, [entity.uuid]
        )

    def bulk_delete(self, entities: List[FolderEntity]):
        entity = entities[0]
        ids = [entity.uuid for entity in entities]
        return self._service.delete_folders(entity.project_id, entity.team_id, ids)

    @staticmethod
    def dict2entity(data: dict) -> FolderEntity:
        try:
            return FolderEntity(
                uuid=data["id"],
                is_root=bool(data["is_root"]),
                team_id=data["team_id"],
                project_id=data["project_id"],
                name=data["name"],
                folder_users=data.get("folder_users"),
            )
        except KeyError:
            raise AppException("Cant serialize folder data")


class AnnotationClassRepository(BaseManageableRepository):
    def __init__(self, service: SuperannotateBackendService, project: ProjectEntity):
        self._service = service
        self.project = project

    def get_one(self, uuid: Condition) -> AnnotationClassEntity:
        raise NotImplementedError

    def get_all(
            self, condition: Optional[Condition] = None
    ) -> List[AnnotationClassEntity]:
        query = condition.build_query() if condition else None
        data = self._service.get_annotation_classes(
            self.project.id, self.project.team_id, query
        )
        if data:
            return [self.dict2entity(data) for data in data]
        return []

    def insert(self, entity: AnnotationClassEntity):
        res = self._service.set_annotation_classes(
            self.project.id, self.project.team_id, [entity]
        )
        if "error" in res:
            raise AppException(res["error"])
        return self.dict2entity(res[0])

    def delete(self, uuid: int):
        self._service.delete_annotation_class(
            team_id=self.project.team_id,
            project_id=self.project.id,
            annotation_class_id=uuid,
        )

    def bulk_insert(self, entities: List[AnnotationClassEntity]):
        res = self._service.set_annotation_classes(
            self.project.id, self.project.team_id, entities
        )
        if "error" in res:
            raise AppException(res["error"])

        return [self.dict2entity(i) for i in res]

    def update(self, entity: AnnotationClassEntity):
        raise NotImplementedError

    @staticmethod
    def dict2entity(data: dict) -> AnnotationClassEntity:
        return AnnotationClassEntity(
            id=data["id"],
            project_id=data["project_id"],
            name=data["name"],
            count=data["count"],
            color=data["color"],
            createdAt=data["createdAt"],
            updatedAt=data["updatedAt"],
            attribute_groups=data["attribute_groups"],
            type=ClassTypeEnum.get_name(data.get("type", 1)),
        )


class ImageRepository(BaseManageableRepository):
    def __init__(self, service: SuperannotateBackendService):
        self._service = service

    def get_one(self, uuid: int) -> ImageEntity:
        raise NotImplementedError

    def get_all(self, condition: Optional[Condition] = None) -> List[ImageEntity]:
        images = self._service.get_images(condition.build_query())
        return [self.dict2entity(image) for image in images]

    def insert(self, entity: ImageEntity) -> ImageEntity:
        raise NotImplementedError

    def delete(self, uuid: int, team_id: int, project_id: int):
        self._service.delete_images(
            image_ids=[uuid], team_id=team_id, project_id=project_id
        )

    def update(self, entity: ImageEntity):
        self._service.update_image(
            image_id=entity.uuid,
            project_id=entity.project_id,
            team_id=entity.team_id,
            data=entity.to_dict(),
        )
        return entity

    @staticmethod
    def dict2entity(data: dict) -> ImageEntity:
        return ImageEntity(
            uuid=data["id"],
            name=data["name"],
            path=data["path"],
            project_id=data["project_id"],
            team_id=data["team_id"],
            annotation_status_code=data["annotation_status"],
            folder_id=data["folder_id"],
            annotator_id=data["annotator_id"],
            annotator_name=data["annotator_name"],
            is_pinned=data.get("is_pinned"),
            created_at=data["createdAt"],
            updated_at=data["updatedAt"],
        )


class UserRepository(BaseReadOnlyRepository):
    @staticmethod
    def dict2entity(data: dict) -> UserEntity:
        return UserEntity(
            uuid=data["id"],
            first_name=data["first_name"],
            last_name=data["last_name"],
            email=data["email"],
            picture=data["picture"],
            user_role=data["user_role"],
        )


class TeamRepository(BaseReadOnlyRepository):
    def __init__(self, service: SuperannotateBackendService):
        self._service = service

    def get_one(self, uuid: int) -> Optional[TeamEntity]:
        res = self._service.get_team(team_id=uuid)
        return self.dict2entity(res)

    def get_all(self, condition: Optional[Condition] = None) -> List[TeamEntity]:
        raise NotImplementedError

    @staticmethod
    def dict2entity(data: dict):
        return TeamEntity(
            uuid=data["id"],
            name=data["name"],
            description=data["description"],
            team_type=data["type"],
            user_role=data["user_role"],
            is_default=data["is_default"],
            users=[UserRepository.dict2entity(user) for user in data["users"]],
            pending_invitations=data["pending_invitations"],
            creator_id=data["creator_id"],
        )


class MLModelRepository(BaseManageableRepository):
    def __init__(self, service: SuperannotateBackendService, team_id: int):
        self._team_id = team_id
        self._service = service

    def get_one(self, uuid: int) -> MLModelEntity:
        raise NotImplementedError

    def get_all(self, condition: Optional[Condition] = None) -> List[MLModelEntity]:
        models = self._service.search_models(condition.build_query())
        return [self.dict2entity(model) for model in models]

    def insert(self, entity: MLModelEntity) -> MLModelEntity:
        data = self._service.start_model_training(self._team_id, entity.to_dict())
        return self.dict2entity(data)

    def delete(self, uuid: int):
        self._service.delete_model(self._team_id, uuid)

    def update(self, entity: MLModelEntity):
        model_data = {k: v for k, v in entity.to_dict().items() if v}
        data = self._service.update_model(
            team_id=self._team_id, model_id=entity.uuid, data=model_data
        )
        return self.dict2entity(data)

    @staticmethod
    def dict2entity(data: dict) -> MLModelEntity:
        return MLModelEntity(
            uuid=data["id"],
            name=data["name"],
            description=data["description"],
            base_model_id=data["base_model_id"],
            model_type=data["type"],
            task=data["task"],
            image_count=data["image_count"],
            path=data["path"],
            config_path=data["config_path"],
            is_global=data["is_global"],
            training_status=data["training_status"],
        )


class IntegrationRepository(BaseReadOnlyRepository):
    def __init__(self, service: SuperannotateBackendService, team_id: int):
        self._service = service
        self._team_id = team_id

    def get_one(self, uuid: int) -> Optional[TeamEntity]:
        raise NotImplementedError

    def get_all(self, condition: Optional[Condition] = None) -> List[IntegrationEntity]:
        return parse_obj_as(
            List[IntegrationEntity], self._service.get_integrations(self._team_id)
        )


class ItemRepository(BaseReadOnlyRepository):
    def __init__(self, service: SuperannotateBackendService):
        self._service = service

    def get_one(self, uuid: Condition) -> Entity:
        items = self._service.list_items(uuid.build_query())
        if len(items) >= 1:
            return Entity(**Entity.map_fields(items[0]))

    def get_all(self, condition: Optional[Condition] = None) -> List[Entity]:
        items = self._service.list_items(condition.build_query())
        return [Entity(**Entity.map_fields(item)) for item in items]
